/* 
   Program generate triangulations by adding faces in lexigraphic order.
*/

/* version 0.20 */

#define USAGE \
" lextri [-v] [-h] [-i] [-o o] [-e e] [-d d] [-g g] [-G G] [-E] [-T] [-F] [-pq p q] [-r res -m mod] nv"

#define HELPTEXT \
" lextri : generate triangulations of surfaces with nv vertices.\n\
\n\
 help switches:\n\
   -h  show this text\n\
   -v  verbose output\n\
\n\
 parameter switches:\n\
  -i       if present only irreducible triangulations are generated\n\
  -o o     0: only nonorientable, 1: only orientable, otherwise: both\n\
  -e e     only equivelar with degree e are generated\n\
  -d d     minimum vertex degree\n\
  -D d     maximum vertex degree\n\
  -g g     minimum Euler genus\n\
  -G G     maximum Euler genus\n\
  -E       all vertices have Even degree\n\
  -B       dual graph is Bipartite\n\
  -T       embedding is vertex Transitive\n\
  -F       embedding is Flag transitive\n\
  -pq p q  generate (triangulation of) equivelar polyhedral maps of type {p,q},\n\
           each triangular face has vertices of degree p, 2q, and 2q\n\
  -r res   res for splitting\n\
  -m mod   mod for splitting\n\
\n\
  nv       number of vertices in generated triangulations.\n"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CPUTIME 1          /* Whether to measure the cpu time or not */

#if CPUTIME
#include <sys/times.h>
#include <time.h>
#if !defined(CLK_TCK) && !defined(_SC_CLK_TCK)
#include <unistd.h>
#endif
#if !defined(CLK_TCK) && defined(_SC_CLK_TCK)
#define CLK_TCK sysconf(_SC_CLK_TCK)
#endif
#if !defined(CLK_TCK) && defined(CLOCKS_PER_SEC)
#define CLK_TCK CLOCKS_PER_SEC
#endif
#if !defined(CLK_TCK)
#define CLK_TCK 60     /* If the CPU time stated by the program appears
                      to be out by a constant ratio, the most likely
                      explanation is that the code got to this point but
                      60 is the wrong guess.  Another common value is 100. */
#endif
#endif

#define FALSE 0
#define TRUE 1

#define MAXN 110
#define MAXF 500
#define MAXG 50

/* BIG macros taken from plantri. */
/* The program is so fast that the count of output graphs can quickly
   overflow a 32-bit integer.  Therefore, we use two long values
   for each count, with a ratio of 10^9 between them.  The macro
   ADDBIG adds a small number to one of these big numbers.  
   BIGTODOUBLE converts a big number to a double (approximately).
   SUMBIGS adds a second big number into a first big number.
   SUBBIGS subtracts a second big number from a first big number.
   PRINTBIG prints a big number in decimal.
   ZEROBIG sets the value of a big number to 0.
   ISZEROBIG tests if the value is 0.
   SETBIG sets a big number to a value at most 10^9-1.
   ISEQBIG tests if two big numbers are equal.
*/

typedef struct
{
    long hi,lo;
} bigint;

#define ZEROBIG(big) big.hi = big.lo = 0L
#define ISZEROBIG(big) (big.lo == 0 && big.hi == 0)
#define SETBIG(big,value) {big.hi = 0L; big.lo = (value);}
#define ADDBIG(big,extra) if ((big.lo += (extra)) >= 1000000000L) \
    { ++big.hi; big.lo -= 1000000000L;}
#define PRINTBIG(file,big) if (big.hi == 0) \
 fprintf(file,"%ld",big.lo); else fprintf(file,"%ld%09ld",big.hi,big.lo)
#define BIGTODOUBLE(big) (big.hi * 1000000000.0 + big.lo)
#define SUMBIGS(big1,big2) {if ((big1.lo += big2.lo) >= 1000000000L) \
    {big1.lo -= 1000000000L; big1.hi += big2.hi + 1L;} \
    else big1.hi += big2.hi;}
#define SUBBIGS(big1,big2) {if ((big1.lo -= big2.lo) < 0L) \
    {big1.lo += 1000000000L; big1.hi -= big2.hi + 1L;} \
    else big1.hi -= big2.hi;}
/* Note: SUBBIGS must not allow the value to go negative.
   SUMBIGS and SUBBIGS both permit big1 and big2 to be the same bigint. */
#define ISEQBIG(big1,big2) (big1.lo == big2.lo && big1.hi == big2.hi)

static int verbose;
static int only_irreducible; /* flag if only irreducible triangulations are to
				be generated */
static int even_degree; /* flag if ever vertex is to have even degree */
static int dual_bipartite; /* flag if dual of triangulation is to be bipartite,
			    faces of triangulation are 2-colorable */
static int vertex_transitive; 
static int vertex_switch; 
                   /* embedding is to be vertex transitive.
		      for any pair of vertices there is automorphism of the 
		      embedded graph which maps one vertex into the other */
static int flag_transitive; 
static int flag_switch; 
                   /* embedding is to be flag transitive.
		      for any pair of flags (vertex, edge, face) there is 
		      automorphism of the 
		      embedded graph which maps vertex onto vertex, edge onto 
		      edge, and face onto face */
static int pq_switch;
                   /* each triangular face has one vertex (face vertex) with 
		      degree pq_degrees[0]=p and two (vertex verices) with 
		      pq_degrees[1]=2q.  */
static int pq_degrees[2];
static int pq_total; /* pq_degrees[0]+2*pq_degrees[1] */
static int nv;     /* number of vertices in final triangulation */
static int ne;     /* number of edges in triangulation */
static int nf;     /* number of faces in triangulation */
static int minnv;  /* minimum number of vertices in a triangulation */
static int maxnv;  /* maximum number of vertices in a triangulation */
static int minnf;  /* minimum number of faces in a triangulation */
static int maxnf;  /* maximum number of faces in a triangulation */
static int ming;   /* minimum Euler genus of a triangulation */
static int maxg;   /* maximum Euler genus of a triangulation */
static int mindegree; /* minimum degree of vertex of a triangulation */
static int maxdegree; /* maximum degree of vertex of a triangulation */
static int edge[MAXN][MAXN]; /* edge[i][j] and edge[j][i] are the vertices of 
				the two faces sharing the edge (i,j).  
				-1 for no corresponding face. */
static int color[MAXN][MAXN]; /* color[i][j] is the color of face 
				 (i,j,edge[i][j]). +1 for first color, -1 for 
				 second color, -999 if color has not yet been 
				 determined.  Undefined if edge[i][j] is -1 or
				 if dual_bipartite is FALSE. */
static int degree[MAXN]; /* edge degree of vertices */
static int fdegree[MAXN]; /* face degree of vertices, number of faces with 
			     that vertex */
static int pq_type[MAXN]; /* defined if pq_switch is TRUE.
			     pq_type[v]=0 if v is a face vertex
			     pq_type[v]=1 if v is a vertex vertex
			     pq_type[v]=-1 if v has not been assigned a type */
static int pq_required[MAXN]; /* if pq_switch then degrees finished vertices 
				 must have */
static int complete[MAXN]; /* link of vertex is a triangulation of a circle */
static int completeness[MAXN]; /* 0 if not complete, 1 if newly complete, or 
				  2 if formerly complete */
static int match[MAXN][2*MAXN][MAXN]; /* lk(v0) is equivalent to lk(0),
				      i is an index of the automorphisms of 
				      this equivalence,
				      v is mapped to match[v0][i][v] by the 
				      i-th automorphism */
static int nmatch[MAXN];   /* number of automorphisms above */
static int list[MAXF][3];  /* list of the vertices in each ordered face */
static bigint ngenerated[MAXN+1][MAXG][2]; 
                              /* number of triangulations generated by number
				 of vertices, Euler genus, and orientability */
static bigint ngenerated_total; /* number of triangulations generated */
static int nt_match;        /* number of triangles matched */
static int v_to_label[MAXN];/* relabeling of vertices */
static int label_to_v[MAXN];/* vertex with label*/
static int next_label;      /* next unused label */
static int debug_count;
static int res,mod;        /* res/mod from command line (default 0/1) */
static char res_text[15];  /* text string for res with padded zeros */
static int splitlevel,
           splitcount;     /* used for res/mod splitting */
static int orientability;  /* orientability of output surfaces, 
			      0 for only nonorientable, 1 for orientable, 
			      otherwise both */
static int equivelar;      /* if > 0, degree for all vertices */
static FILE *outfile_alpha[MAXN][MAXG][2];
static FILE *outfile_lex[MAXN][MAXG][2];

#define MAX(x,y) ((x)<(y) ? (y) : (x))
#define MIN(x,y) ((x)>(y) ? (y) : (x))

#ifdef SPLITTEST
static bigint splitcases;
#endif

void next_face();

static char i2a[52] = {
  'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
  'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'};

void error_exit(int errornum)
{
  fprintf(stderr,"Error number %d, exiting\n",errornum);
  exit(1);
}

void write_alpha(FILE *output)
{
  /* write out surface ala plantri format */

  int v0,v1,v2,v3,i;
  int v11[MAXN],v22[MAXN];

  for (v0=0; v0<nv; v0++)
    v11[v0] = -1;
  v0 = 0;
  for (v11[v0]=1; edge[v0][v11[v0]] == -1; v11[v0]++) {}
  v22[v0] = edge[v0][v11[v0]];

  fprintf(output,"%d ",nv);
  for (v0=0; v0<nv; v0++) {
    v1 = v11[v0];
    v2 = v22[v0];
    for (i=0; i<degree[v0]; i++) {
      fprintf(output,"%c",v1+'a');
      if (v11[v1] == -1) {
	v11[v1] = v0;
	if (edge[v1][v0] == v2)
	  v22[v1] = edge[v0][v1];
	else
	  v22[v1] = edge[v1][v0];
      }
      v3 = v1;
      v1 = v2;
      if (edge[v0][v1] == v3)
	v2 = edge[v1][v0];
      else
	v2 = edge[v0][v1];
    }
    if (v0 < nv-1)
      fprintf(output,",");
    else
      fprintf(output,"\n");
  }
}

void write_alpha_pq(FILE *output)
{
  /* write out surface ala plantri format */
  /* pq_switch is TRUE.  write out just the equivelar polyhedral map. */

  int p_nv,p_nf,p_v;
  int v0,v1,v2,v3,i;
  int v11[MAXN],v22[MAXN];
  int v_to_label[MAXN],label_to_v[MAXN],nlabel;
  int first_nei;

  for (v0=0; v0<nv; v0++)
    v11[v0] = -1;
  if (pq_type[0] == 1)
    v0 = 0;
  else
    v0 = 1;
  nlabel = 0;
  label_to_v[nlabel] = v0;
  v_to_label[v0] = nlabel++;
  for (v11[v0]=1; edge[v0][v11[v0]] == -1; v11[v0]++) {}
  v22[v0] = edge[v0][v11[v0]];

  p_nf = nf/pq_degrees[0];
  p_nv = nv - p_nf;

  if (FALSE) {
    fprintf(output,"# %d ",p_nv);
    for (p_v=0; p_v<p_nv; p_v++) {
      v0 = label_to_v[p_v];
      v1 = v11[v0];
      v2 = v22[v0];
      for (i=0; i<degree[v0]; i++) {
	if (pq_type[v1] == 1) {
	  if (v11[v1] == -1) {
	    label_to_v[nlabel] = v1;
	    v_to_label[v1] = nlabel++;
	    v11[v1] = v0;
	    if (edge[v1][v0] == v2)
	      v22[v1] = edge[v0][v1];
	    else
	      v22[v1] = edge[v1][v0];
	  }
	  fprintf(output,"%c",v_to_label[v1]+'a');
	}
	v3 = v1;
	v1 = v2;
	if (edge[v0][v1] == v3)
	  v2 = edge[v1][v0];
	else
	  v2 = edge[v0][v1];
      }
      if (p_v < p_nv-1)
	fprintf(output,",");
      else
	fprintf(output,"\n");
    }
  }
  if (TRUE) {
    for (p_v=0; p_v<p_nv; p_v++) {
      if (p_v == 0)
	fprintf(output,"G = {%d:[",p_v);
      else
	fprintf(output,"%d:[",p_v);

      v0 = label_to_v[p_v];
      v1 = v11[v0];
      v2 = v22[v0];
      first_nei = TRUE;
      for (i=0; i<degree[v0]; i++) {
	if (pq_type[v1] == 1) {
	  if (v11[v1] == -1) {
	    label_to_v[nlabel] = v1;
	    v_to_label[v1] = nlabel++;
	    v11[v1] = v0;
	    if (edge[v1][v0] == v2)
	      v22[v1] = edge[v0][v1];
	    else
	      v22[v1] = edge[v1][v0];
	  }
	  if (!first_nei)
	    fprintf(output,",");
	  first_nei = FALSE;
	  fprintf(output,"%d",v_to_label[v1]);
	}
	v3 = v1;
	v1 = v2;
	if (edge[v0][v1] == v3)
	  v2 = edge[v1][v0];
	else
	  v2 = edge[v0][v1];
      }
      if (p_v < p_nv-1)
	fprintf(output,"],");
      else 
	fprintf(output,"]}\nprint(G)\nprint(isPartialCube(G))\n");
    }
  }
}

void write_lex(FILE *output)
{
  /* write out surface in sortable format.  one line of text for each surface.
     one character for each vertex of each face. */

  int iface;

  for (iface=0; iface<nf; iface++)
    //    fprintf(output,"%c%c%c",
    //    list[iface][0]+'a',list[iface][1]+'a',list[iface][2]+'a');
    fprintf(output,"%c%c%c",
	    i2a[list[iface][0]],i2a[list[iface][1]],i2a[list[iface][2]]);
  fprintf(output,"\n");
}

void dump_it()
{
  int ilist;
  int v0,v1;

  fprintf(stderr,"nv = %d, ne = %d, nf = %d\n",nv,ne,nf);

  fprintf(stderr,"list of faces:\n");
  for (ilist=0; ilist<nf; ilist++)
    fprintf(stderr,"%2d %2d %2d \n",
	    list[ilist][0],list[ilist][1],list[ilist][2]);

  fprintf(stderr,"edge matrix:\n");
  fprintf(stderr,"   |");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",v1);
  fprintf(stderr,"\n");
  fprintf(stderr,"----");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"---");
  fprintf(stderr,"\n");
  for (v0=0; v0<nv; v0++) {
    fprintf(stderr,"%2d |",v0);
    for (v1=0; v1<nv; v1++) 
      if (edge[v0][v1] != -1)
	fprintf(stderr,"%2d ",edge[v0][v1]);
      else
	fprintf(stderr,"   ");
    fprintf(stderr,"\n");
  }

  fprintf(stderr,"edge degree:");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",degree[v1]);
  fprintf(stderr,"\n");

  fprintf(stderr,"face degree:");
  for (v1=0; v1<nv; v1++) 
    fprintf(stderr,"%2d ",fdegree[v1]);
  fprintf(stderr,"\n");

}

void add_face(int v0, int v1, int v2, int lcolor)
{
  if (lcolor != -999 && lcolor != +1 && lcolor != -1)
    error_exit(999);
  if (0 > v0)
    error_exit(5);

  if (v0 >= v1)
    error_exit(6);

  if (v1 >= v2)
    error_exit(7);

  if (v2 > nv)
    error_exit(8);

  if (edge[v0][v1] != -1 &&
      edge[v1][v0] != -1)
    error_exit(1);
  if (edge[v1][v2] != -1 &&
      edge[v2][v1] != -1)
    error_exit(2);
  if (edge[v2][v0] != -1 &&
      edge[v0][v2] != -1)
    error_exit(3);

  if (nf >= MAXF)
    error_exit(4);
  list[nf][0] = v0;
  list[nf][1] = v1;
  list[nf][2] = v2;
  nf++;

  if (v2 == nv) {
    nv++;
    if (pq_switch) {
      pq_type[v2] = 2 -(pq_type[v0] + pq_type[v1]);
      pq_required[v2] = pq_degrees[pq_type[v2]];
    }
  }

  if (edge[v0][v1] == -1) {
    edge[v0][v1] = v2;
    color[v0][v1] = lcolor;
    if (edge[v1][v0] == -1) {
      degree[v0]++;
      degree[v1]++;
      ne++;
    }
  }
  else {
    edge[v1][v0] = v2;
    color[v1][v0] = lcolor;
  }
 
  if (edge[v1][v2] == -1) {
    edge[v1][v2] = v0;
    color[v1][v2] = lcolor;
    if (edge[v2][v1] == -1) {
      degree[v1]++;
      degree[v2]++;
      ne++;
    }
  }
  else {
    edge[v2][v1] = v0;
    color[v2][v1] = lcolor;
  }
 
  if (edge[v2][v0] == -1) {
    edge[v2][v0] = v1;
    color[v2][v0] = lcolor;
    if (edge[v0][v2] == -1) {
      degree[v2]++;
      degree[v0]++;
      ne++;
    }
  }
  else {
    edge[v0][v2] = v1;
    color[v0][v2] = lcolor;
  }

  fdegree[v0]++;
  fdegree[v1]++;
  fdegree[v2]++;

  /* update complete */
  
  if (degree[v0] == fdegree[v0]) {
    complete[v0] = TRUE;
    completeness[v0] = 1;
  }
  if (degree[v1] == fdegree[v1]) {
    complete[v1] = TRUE;
    completeness[v1] = 1;
  }
  if (degree[v2] == fdegree[v2]) {
    complete[v2] = TRUE;
    completeness[v2] = 1;
  }
}

void remove_face()
{
  int v0,v1,v2;

  nf--;
  v0 = list[nf][0];
  v1 = list[nf][1];
  v2 = list[nf][2];

  complete[v0] = FALSE;
  complete[v1] = FALSE;
  complete[v2] = FALSE;
  completeness[v0] = 0;
  completeness[v1] = 0;
  completeness[v2] = 0;

  if (edge[v0][v1] == v2) {
    edge[v0][v1] = -1;
    if (edge[v1][v0] == -1) {
      degree[v0]--;
      degree[v1]--;
      ne--;
    }
  }
  else {
    edge[v1][v0] = -1;
    if (edge[v0][v1] == -1) {
      degree[v0]--;
      degree[v1]--;
      ne--;
    }
  }
    
  if (edge[v1][v2] == v0) {
    edge[v1][v2] = -1;
    if (edge[v2][v1] == -1) {
      degree[v1]--;
      degree[v2]--;
      ne--;
    }
  }
  else {
    edge[v2][v1] = -1;
    if (edge[v1][v2] == -1) {
      degree[v1]--;
      degree[v2]--;
      ne--;
    }
  }
    
  if (edge[v2][v0] == v1) {
    edge[v2][v0] = -1;
    if (edge[v0][v2] == -1) {
      degree[v2]--;
      degree[v0]--;
      ne--;
    }
  }
  else {
    edge[v0][v2] = -1;
    if (edge[v2][v0] == -1) {
      degree[v2]--;
      degree[v0]--;
      ne--;
    }
  }

  fdegree[v0]--;
  fdegree[v1]--;
  fdegree[v2]--;
  if (fdegree[v2] == 0) {
    nv--;
    if (pq_switch) {
      pq_type[v2] = -1;
      pq_required[v2] = -1;
    }
  }
}

void initialize()
{
  /* one time initialization */

  int orient,v,g;

#ifdef SPLITTEST
    ZEROBIG(splitcases);
#endif

  debug_count = 0;

  for (v=0; v<=MAXN; v++)
    for (g=0; g<MAXG; g++)
      for (orient=0; orient<2; orient++)
	ZEROBIG(ngenerated[v][g][orient]);
  ZEROBIG(ngenerated_total);
}

void initialize_first_vertex(int degree0, int pq_type0, int pq_type1)

{

  int i,j;
  int lcolor;

  /* clear data*/

  for (i=0; i<maxnv; i++)
    for (j=0; j<maxnv; j++)
      edge[i][j] = -1;
  for (i=0; i<maxnv; i++) {
    degree[i] = 0;
    fdegree[i] = 0;
    v_to_label[i] = -1;
    complete[i] = FALSE;
  }

  if (pq_switch) {
    pq_type[0] = pq_type0;
    pq_required[0] = pq_degrees[pq_type0];
    pq_type[1] = pq_type1;
    pq_required[1] = pq_degrees[pq_type1];
    for (i=2; i<maxnv; i++) {
      pq_type[i] = -1;
      pq_required[i] = -1;
    }
  }

  nv = 2;
  ne = 0;
  nf = 0;

  /* add faces around first vertex */

  lcolor = +1;
  add_face(0,1,2,lcolor);
  for (i=3; i<=degree0; i++) {
    if (i%2==1)
      lcolor = -lcolor;
    add_face(0,i-2,i,lcolor);
  }
  lcolor = -lcolor;
  add_face(0,degree0-1,degree0,lcolor);

} 

int orientable()
{
  /* check if surface is orientable. */
  
  /* only the strongly connected component containing the first face is 
     checked.
     false positive is possible. */
  
  int orient[MAXN][MAXN];  
                   /* orient[i][j] is orientation of face (i,j,edge[i][j]) . 
		      +1 if clockwise,
		      -1 if counter-clockwise,
		      0 if don't know yet. */
  int iface;         /* face index */
  int v0,v1,v2;
  int face_orient[3];
  int up,down,new_orient;
  int ofound[MAXF]; /* orientation of face has been determined */
  int nfound;       /* number of faces for which orientation has been 
		       determined */
  int more_found;
  
  for (v0=0; v0<nv; v0++)
    for (v1=0; v1<nv; v1++)
      orient[v0][v1] = 0;
  for (iface=0; iface<nf; iface++)
    ofound[iface] = FALSE;
  nfound = 0;
  
  if (edge[list[0][0]][list[0][1]] == list[0][2])
    orient[list[0][0]][list[0][1]] = +1;
  else
    orient[list[0][1]][list[0][0]] = -1;
  
  if (edge[list[0][1]][list[0][2]] == list[0][0])
    orient[list[0][1]][list[0][2]] = +1;
  else
    orient[list[0][2]][list[0][1]] = -1;
  
  if (edge[list[0][2]][list[0][0]] == list[0][1])
    orient[list[0][2]][list[0][0]] = +1;
  else
    orient[list[0][0]][list[0][2]] = -1;
  
  ofound[0] = TRUE;
  nfound++;
  
  more_found = TRUE;
  while (nfound < nf && more_found) {
    more_found = FALSE;
    for (iface = 1; iface<nf; iface++) 
      if (!ofound[iface]) {
	v0 = list[iface][0];
	v1 = list[iface][1];
	v2 = list[iface][2];
	
	if (edge[v0][v1] != v2)
	  face_orient[0] = -orient[v0][v1];
	else
	  face_orient[0] = orient[v1][v0];
	
	if (edge[v1][v2] != v0)
	  face_orient[1] = -orient[v1][v2];
	else
	  face_orient[1] = orient[v2][v1];
	
	if (edge[v2][v0] != v1)
	  face_orient[2] = -orient[v2][v0];
	else
	  face_orient[2] = orient[v0][v2];
	
	down = (face_orient[0] == -1 ||
		face_orient[1] == -1 ||
		face_orient[2] == -1);
	up = (face_orient[0] == +1 ||
	      face_orient[1] == +1 ||
	      face_orient[2] == +1);
	
	if (up && down)
	  return FALSE;
	
	if (up || down) {
	  more_found = TRUE;
	  ofound[iface] = TRUE;
	  nfound++;
	  if (down)
	    new_orient = -1;
	  else
	    new_orient = +1;
	  
	  if (edge[v0][v1] == v2)
	    orient[v0][v1] = new_orient;
	  else
	    orient[v1][v0] = -new_orient;
	  
	  if (edge[v1][v2] == v0)
	    orient[v1][v2] = new_orient;
	  else
	    orient[v2][v1] = -new_orient;
	  
	  if (edge[v2][v0] == v1)
	    orient[v2][v0] = new_orient;
	  else
	    orient[v0][v2] = -new_orient;
	}
      }
  } 
  return TRUE;
}

int admissable(int v0, int v1, int v2)
{
  /* check if adding face (v0,v1,v2) would cause the link of v0 to be a cycle 
     and at least one other edge */
  
  int v3,v4,v5,v6;

  if (fdegree[v0] < 2)
    return TRUE;

  if (equivelar > 0) {
    int new_degree;
    
    if (degree[v0] + 2 > equivelar) {
      new_degree = degree[v0];
      if (edge[v0][v1] == -1 && edge[v1][v0] == -1)
	new_degree++;
      if (edge[v0][v2] == -1 && edge[v2][v0] == -1)
	new_degree++;
      if (new_degree > equivelar)
	return FALSE;
    }
  }

  if (pq_switch) {
    int new_degree;
    
    if (degree[v0] + 2 > pq_required[v0]) {
      new_degree = degree[v0];
      if (edge[v0][v1] == -1 && edge[v1][v0] == -1)
	new_degree++;
      if (edge[v0][v2] == -1 && edge[v2][v0] == -1)
	new_degree++;
      if (new_degree > pq_required[v0])
	return FALSE;
    }
  }    
  
  
  if (degree[v0] == fdegree[v0]+1)
    return TRUE;
  
  if (edge[v0][v1] == -1) {
    if (edge[v1][v0] == -1)
      return TRUE;
    else
      v3 = edge[v1][v0];
  }
  else
    v3 = edge[v0][v1];
  
  if (edge[v0][v2] == -1) {
    if (edge[v2][v0] == -1)
      return TRUE;
    else
      v4 = edge[v2][v0];
  }
  else
    v4 = edge[v0][v2];
  
  v5 = v1;
  while (v3 != v4 && v3 != -1) {
    if (edge[v0][v3] == v5)
      v6 = edge[v3][v0];
    else
      v6 = edge[v0][v3];
    v5 = v3;
    v3 = v6;
  }

  if (v3 == v4)
    return FALSE;
  else
    return TRUE;
}

int smaller_lex()
{
  /* check if the relabeling starting at nt_match is lexigraphically 
     smaller than the current labeling (prune if so) */
  /* if vertex_transitive is TRUE then prune also if relabeling is larger */
  /* return 1 if smaller, -1 if larger, 0 if same (so far) */
  
  int label[3];         /* vertices of face. also labels in relabeling */
  int w0,w1,w2;         /* vertices of face relabeled 
			   (label[0],label[1],label[2]) */
  int i;                /* vertex index for face */
  int branch;           /* search needs to branch */
  int retval1,retval2;  /* values returned by first and second branch  */
  int save_next_label;
  int save_nt_match;
  int retval;

  branch = FALSE;

  for (i=0; i<3; i++)
    label[i] = list[nt_match-1][i];
  
  while (nt_match<nf) {
    
    /* label[0], label[1], and label[2] are the vertices of the nt_match-1 or 
       better */
    /* find the next relabeled face */
    
    w2 = -1;
    while (w2 == -1) {
      
      w0 = label_to_v[label[0]];
      w1 = label_to_v[label[1]];
      
      if (edge[w0][w1] != -1 &&
	  edge[w1][w0] != -1) {
	if (v_to_label[edge[w0][w1]] > label[2] &&
	    v_to_label[edge[w1][w0]] > label[2]) {
	  if (v_to_label[edge[w0][w1]] > v_to_label[edge[w1][w0]])
	    w2 = edge[w1][w0];
	  else
	    w2 = edge[w0][w1];
	}
	else if (v_to_label[edge[w0][w1]] > label[2])
	  w2 = edge[w0][w1];
	else if (v_to_label[edge[w1][w0]] > label[2])
	  w2 = edge[w1][w0];
	else if (v_to_label[edge[w0][w1]] == -1) {
	  
	  if (v_to_label[edge[w1][w0]] == -1)
	    
	    /* both 3rd vertices of the faces of (w0,w1) are unlabeled.  
	       the search must branch below */
	    
	    branch = TRUE;
	  
	  w2 = edge[w0][w1];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
	else if (v_to_label[edge[w1][w0]] == -1) {
	  w2 = edge[w1][w0];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      else if (edge[w0][w1] != -1) {
	if (v_to_label[edge[w0][w1]] > label[2])
	  w2 = edge[w0][w1];
	else if (v_to_label[edge[w0][w1]] == -1) {
	  w2 = edge[w0][w1];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      else if (edge[w1][w0] != -1) {
	if (v_to_label[edge[w1][w0]] > label[2])
	  w2 = edge[w1][w0];
	else if (v_to_label[edge[w1][w0]] == -1) {
	  w2 = edge[w1][w0];
	  v_to_label[w2] = next_label;
	  label_to_v[next_label] = w2;
	  next_label++;
	}
      }
      
      if (w2 == -1) {
	if (label[0] == list[nt_match][0] && label[1] == list[nt_match][1])
	  
	  /* current labeling has another face with (label[0],label[1]) but
	     relabeling does not. */
	  
	  {
	    if (complete[w0] && complete[w1]) {
	      
	      /* this relabeling is larger */
	      
		return -1;
	    }
	    else
	      
	      /* this relabeling is the same so far */
	      
	      return 0;
	  }
	else {
	  
	  /* neither current labeling nor relabeling have another face with 
	     (label[0],label[1]). */
	  
	  label[1]++;
	  if (label[1] == next_label || label[1] >= nv-1) {
	    label[0]++;
	    label[1] = label[0] + 1;
	  }
	  label[2] = label[1];
	}
      }
      else {  /* w2 != -1 */
	label[2] = list[nt_match][2];
	if (label[0] == list[nt_match][0] && label[1] == list[nt_match][1]) {
	  if (v_to_label[w2] > label[2]) {
	    if (edge[w0][w1] != -1 && edge[w1][w0] != -1)
	      
	      /* current labeling and relabeling both have another face with 
		 (label[0],label[1]).  current labeling is smaller. */
	      
	      return -1;
	    
	    else
	      
	      /* current labeling has another face with 
		 (label[0],label[1]) but relabeling does not.  */

	      return 0;

	  }
	  
	  else if (v_to_label[w2] < label[2])
	    
	    /* current labeling and relabeling both have another face with 
	       (label[0],label[1]).  relabeling is smaller. */
	    
	    return 1;
	  
	  /* current labeling and relabeling both have another face with 
	     (label[0],label[1]).  they are the same. */
	  
	  if (branch) {

	    /* relabeling has two possible ways to order the faces with 
	       (w0,w1).  this is the first of the two possibilities. */

	    save_nt_match = nt_match;
	    nt_match++;
	    save_next_label = next_label-1;
	    retval1 = smaller_lex();
	    if (retval1 == 1)
	      return retval1;
	    for ( ; next_label>save_next_label; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    nt_match = save_nt_match;
	    v_to_label[w2] = -1;

	    /* this is the second possibility */

	    w2 = edge[w1][w0];
	    v_to_label[w2] = next_label;
	    label_to_v[next_label] = w2;
	    next_label++;
	    nt_match++;
	    retval2 = smaller_lex();
	    if (retval2 == 1)
	      return retval2;
	    
	    if (retval1 == 0 || retval2 == 0)
	      return 0;
	    else
	      return -1;
	  }

	  /* drop out of while loop */
	  
	}
	else
	  
	  /* current labeling does not have another face with 
	     (label[0],label[1]) but relabeling does. */
	  
	  return 1;
      }
      
    }
    nt_match++;
  }
  
  /* they are the same */
  
  return 0;
}

int min_lex()
{
  /* check that the current labeling of the list of faces is lexigraphical
     minimum */
  /* if vertex_transitive is TRUE then check that current labeling is the same
     as at least one labeling per vertex. */
  /* if flag_transitive is TRUE then check that current labeling is the same
     as all labeling others. */

  int v0,v1,v;
  int ineighbor,ioffset,save;
  int neighbor[MAXN];
  int save_next_label;
  int imatch;
  int retval;
  int trans[MAXN];   /* number of relabelings which are the same so far */
  int nontrans[MAXN];/* number of relabelings which are the different so far */

  save_next_label = degree[0]+1;

  for (v0=0; v0<nv; v0++) {
    if (completeness[v0] == 1)
      nmatch[v0] = 0;
    vertex_transitive = vertex_switch;
    flag_transitive = flag_switch;
    trans[v0] = nontrans[v0] = 0;
    if (completeness[0] != 2 || completeness[v0] != 2) {
      if (degree[v0] == degree[0] &&  complete[v0]) {
	
	/* make a list of the neighbors of v0 */
	
	for (v1=0; edge[v0][v1] == -1; v1++) {}
	neighbor[0] = v1;
	neighbor[1] = edge[v0][v1];
	for (ineighbor=2; ineighbor<degree[0]; ineighbor++)
	  if (edge[v0][neighbor[ineighbor-1]] == neighbor[ineighbor-2])
	    neighbor[ineighbor] = edge[neighbor[ineighbor-1]][v0];
	  else
	    neighbor[ineighbor] = edge[v0][neighbor[ineighbor-1]];
	
	/* set the initial values of v_to_label for v0 */
	
	v_to_label[v0] = 0;
	label_to_v[0] = v0;
	v_to_label[v1] = 1;
	label_to_v[1] = v1;
	for (ioffset=1; ioffset<=degree[0]/2; ioffset++) {
	  v_to_label[neighbor[ioffset]] = 2*ioffset;
	  label_to_v[2*ioffset] = neighbor[ioffset];
	}
	for (ioffset=1; ioffset<(degree[0]+1)/2; ioffset++) {
	  v_to_label[neighbor[-ioffset+degree[0]]] = 2*ioffset+1;
	  label_to_v[2*ioffset+1] = neighbor[-ioffset+degree[0]];
	}
	for (ineighbor=0; ineighbor<degree[0]; ineighbor++) {
	  
	  /* check starting at each neighbor of v0 going clockwise */
	  
	  if (pq_switch) {
	    if (flag_switch)
	      flag_transitive = 
		(pq_type[0] == pq_type[v0] && 
		 pq_type[1] == pq_type[label_to_v[1]]);
	    else if (vertex_switch)
	      vertex_transitive = 
		(pq_type[0] == pq_type[v0] && 
		 pq_type[1] == pq_type[label_to_v[1]]);
	  }
	  nt_match = fdegree[v0];
	  next_label = save_next_label;
	  retval = smaller_lex();
	  if (retval == 1 ||
	      (flag_transitive && retval == -1)) {
	    for ( ; next_label>0; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    return FALSE;
	  }
	  if (vertex_transitive) {
	    if (retval == 0)
	      trans[v0]++;
	    else
	      nontrans[v0]++;
	  }
	  for ( ; next_label>save_next_label; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  if (completeness[v0] == 1)
	    if (nt_match >= degree[0]) {
	      for (v=0; v<=degree[0]; v++)
		match[v0][nmatch[v0]][v] = label_to_v[v];
	      nmatch[v0]++;
	    }
	  
	  
	  /* rotate the labeling for next neighbor */
	  
	  save = v_to_label[neighbor[degree[0]-1]];
	  for (ioffset=degree[0]-1; ioffset>0; ioffset--) {
	    v_to_label[neighbor[ioffset]] = v_to_label[neighbor[ioffset-1]];
	    label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  }
	  v_to_label[neighbor[0]] = save;
	  label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	}
	
	/* flip the labeling for counterclockwise check */
	
	for (ioffset=1; ioffset<(degree[0]+1)/2; ioffset++) {
	  save = v_to_label[neighbor[ioffset]];
	  v_to_label[neighbor[ioffset]] = 
	    v_to_label[neighbor[-ioffset+degree[0]]];
	  label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  v_to_label[neighbor[-ioffset+degree[0]]] = save;
	  label_to_v[v_to_label[neighbor[-ioffset+degree[0]]]] = 
	    neighbor[-ioffset+degree[0]];
	}
	
	for (ineighbor=0; ineighbor<degree[0]; ineighbor++) {
	  
	  /* check starting at each neighbor of v0 going counterclockwise */
	  
	  if (pq_switch) {
	    if (flag_switch)
	      flag_transitive = 
		(pq_type[0] == pq_type[v0] && 
		 pq_type[1] == pq_type[label_to_v[1]]);
	    else if (vertex_switch)
	      vertex_transitive = 
		(pq_type[0] == pq_type[v0] && 
		 pq_type[1] == pq_type[label_to_v[1]]);
	  }
	  nt_match = fdegree[0];
	  next_label = save_next_label;
	  retval = smaller_lex();
	  if (retval == 1 ||
	      (flag_transitive && retval == -1)) {
	    for ( ; next_label>0; next_label--)
	      v_to_label[label_to_v[next_label-1]] = -1;
	    return FALSE;
	  }
	  if (vertex_transitive) {
	    if (retval == 0)
	      trans[v0]++;
	    else
	      nontrans[v0]++;
	  }
	  for ( ; next_label>save_next_label; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  if (completeness[v0] == 1)
	    if (nt_match >= degree[0]) {
	      for (v=0; v<=degree[0]; v++)
		match[v0][nmatch[v0]][v] = label_to_v[v];
	      nmatch[v0]++;
	    }
	  
	  /* rotate the labeling for next neighbor */
	  
	  save = v_to_label[neighbor[degree[0]-1]];
	  for (ioffset=degree[0]-1; ioffset>0; ioffset--) {
	    v_to_label[neighbor[ioffset]] = v_to_label[neighbor[ioffset-1]];
	    label_to_v[v_to_label[neighbor[ioffset]]] = neighbor[ioffset];
	  }
	  v_to_label[neighbor[0]] = save;
	  label_to_v[v_to_label[neighbor[0]]] = neighbor[0];
	}
	
	/* undo labeling before next v0 or returning */
	
	for ( ; next_label>0; next_label--)
	  v_to_label[label_to_v[next_label-1]] = -1;
      }
    }
    
    else if (nmatch[v0] != 0) {
      
      /* completeness[0] == 2 && completeness[v0] == 2 
	 and lk(0) is equivalent to lk(v0) */
      
      for (imatch=0; imatch<nmatch[v0]; imatch++) {
	for (v=0; v<=degree[0]; v++) {
	  label_to_v[v] = match[v0][imatch][v];
	  v_to_label[label_to_v[v]] = v;
	}
	if (pq_switch) {
	  if (flag_switch)
	    flag_transitive = 
	      (pq_type[0] == pq_type[v0] && 
	       pq_type[1] == pq_type[label_to_v[1]]);
	  else if (vertex_switch)
	    vertex_transitive = 
	      (pq_type[0] == pq_type[v0] && 
	       pq_type[1] == pq_type[label_to_v[1]]);
	}
	nt_match = fdegree[0];
	next_label = degree[0] + 1;
	retval = smaller_lex();
	if (retval == 1 ||
	    (flag_transitive && retval == -1)) {
	  for ( ; next_label>0; next_label--)
	    v_to_label[label_to_v[next_label-1]] = -1;
	  return FALSE;
	}
	if (vertex_transitive) {
	  if (retval == 0)
	    trans[v0]++;
	  else
	    nontrans[v0]++;
	}
	for ( ; next_label>degree[0]+1; next_label--)
	  v_to_label[label_to_v[next_label-1]] = -1;
      }
      for ( ; next_label>0; next_label--)
	v_to_label[label_to_v[next_label-1]] = -1;
    }
    if (complete[0] && completeness[v0] == 1)
      completeness[v0] = 2;
    if (vertex_switch && nontrans[v0] != 0 && trans[v0] == 0)
      return FALSE;
  }
  
  //fprintf(stderr,"nv %d\n",nv);

  return TRUE;
}

int contractible(int v0, int v1)
{
  /* check to see if the edge (v0,v1) is contractible.  vertices v0 and v1 are
     complete. */

  int v;
  int neighbors;  /* number of common neighbors of v0 and v1 */

  neighbors = 0;
  for (v=0; v<nv; v++)
    if (edge[v0][v] != -1 && edge[v1][v] != -1) {
      neighbors++;
      if (neighbors == 3)
	return FALSE;
    }
  return TRUE;
}

void found_one(int g, int orient)
{
  /* found a surface */

  char outfilename[100];

  ADDBIG(ngenerated[nv][g][orient],1);
  ADDBIG(ngenerated_total,1);
  
  if (FALSE) {
  if (!pq_switch) {
    if (outfile_alpha[nv][g][orient] == NULL) {
      sprintf(outfilename,"2-manifolds_v%d_o%d_g%d_r%s_m%d.alpha%c",
	      nv,orient,g,res_text,mod,0);
      if ((outfile_alpha[nv][g][orient] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    
    write_alpha(outfile_alpha[nv][g][orient]);
  }
 
  else {
    if (outfile_alpha[nv][g][orient] == NULL) {
      sprintf(outfilename,"2-manifolds_v%d_p%d_q%d_o%d_g%d_r%s_m%d.alpha%c",
	      nv,pq_degrees[0],pq_degrees[1]/2,orient,g,res_text,mod,0);
      if ((outfile_alpha[nv][g][orient] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    
    write_alpha_pq(outfile_alpha[nv][g][orient]);
  }
  }
 
  if (TRUE) {
    if (outfile_lex[nv][g][orient] == NULL) {
      if (equivelar > 0) {
	if (orient)
	  sprintf(outfilename,"equi_surfaces_lex_n%d_p3_q%d_o%d_g%d_r%s_m%d.lex%c",
		  nv,equivelar,orient,g/2,res_text,mod,0);
	else
	  sprintf(outfilename,"equi_surfaces_lex_n%d_p3_q%d_o%d_g%d_r%s_m%d.lex%c",
		  nv,equivelar,orient,g,res_text,mod,0);
      }
      else {
	sprintf(outfilename,"2-manifolds_v%d_o%d_g%d_r%s_m%d.lex%c",
		nv,orient,g,res_text,mod,0);
      }
      if ((outfile_lex[nv][g][orient] = fopen(outfilename,"w")) == NULL)
	{
	  fprintf(stderr,
		  "can't open %s for writing\n",outfilename);
	  exit(1);
	}
    }
    
    write_lex(outfile_lex[nv][g][orient]);
    fflush(outfile_lex[nv][g][orient]);

    //fprintf(stderr,"found one, nv = %d\n",nv);

  }

}

void do_face(int v0, int v1, int v2)
{
  int done;
  int v,g,orient;
  int nattached;
  int lcolor,ocolor,uncolored,fcolor[MAXF],iface;

  /* add a new face and check pruning conditions. */

  /* pruning before adding face */

  if (v2 == maxnv)
    return;

  if (nf == maxnf)
    return;

  /* check for required degree conflict */

  if (pq_switch) {
    if (pq_type[v2] != -1) {
      if (pq_type[v0]+pq_type[v1]+pq_type[v2] != 2)
	return;
    }
    else {
      if (pq_type[v0]+pq_type[v1] == 0)
	return;
    }
  }

  /* check that adding face would not cause the link of v0, v1, or v2 to be a 
     cycle and at least one other edge.
     also check that new degrees of v0, v1, and v2 would not exceed bounds
     required by equivelar or pq_required[] */

  if (!admissable(v0,v1,v2))
    return;
  if (!admissable(v1,v2,v0))
    return;
  if (!admissable(v2,v0,v1))
    return;

  /* if checking for bipartite dual then determine color of new face */
  /* lcolor is the color of the neighbors of the face to be added */

  lcolor = -999;
  if (dual_bipartite) {
    if (edge[v0][v1] != -1)
      ocolor = color[v0][v1];
    else if (edge[v1][v0] != -1)
      ocolor = color[v1][v0];
    else
      ocolor = -999;
    if (ocolor == -lcolor)
      return;
    if (ocolor != -999)
      lcolor = ocolor;

    if (edge[v1][v2] != -1)
      ocolor = color[v1][v2];
    else if (edge[v2][v1] != -1)
      ocolor = color[v2][v1];
    else
      ocolor = -999;
    if (ocolor == -lcolor)
      return;
    if (ocolor != -999)
      lcolor = ocolor;

    if (edge[v2][v0] != -1)
      ocolor = color[v2][v0];
    else if (edge[v0][v2] != -1)
      ocolor = color[v0][v2];
    else
      ocolor = -999;
    if (ocolor == -lcolor)
      return;
    if (ocolor != -999)
      lcolor = ocolor;


    if (lcolor != -999)
      lcolor = -lcolor;
  }

  /* add a new face */

  add_face(v0,v1,v2,lcolor);

  /* check if degree of newly completed vertex do not match pq_required[] */

  if (pq_switch)
    if ((complete[v0] && degree[v0] != pq_required[v0]) ||
	(complete[v1] && degree[v1] != pq_required[v1]) ||
	(complete[v2] && degree[v2] != pq_required[v2]) ) {
      remove_face();
      return;
    }
  
  /* check that vertex 0 still has minimal degree */
  
  if ((degree[v0] < degree[0] && complete[v0]) ||
      (degree[v1] < degree[0] && complete[v1]) ||
      (degree[v2] < degree[0] && complete[v2]) ) {
    remove_face();
    return;
  }

  /* check that all recently completed vertex is even, if required */

  if (even_degree)
    if ((complete[v0] && degree[v0]%2 != 0) ||
	(complete[v1] && degree[v1]%2 != 0) ||
	(complete[v2] && degree[v2]%2 != 0)) {
      remove_face();
      return;
    }
  
  /* check for correct orientability */

  if (orientability == 1) {

    /* don't bother if new face is attached by no or one edge or
       if attached by two edges and completes the common vertex */

    nattached = -3;
    if (edge[v0][v1] != -1) nattached++;
    if (edge[v0][v2] != -1) nattached++;
    if (edge[v1][v0] != -1) nattached++;
    if (edge[v1][v2] != -1) nattached++;
    if (edge[v2][v0] != -1) nattached++;
    if (edge[v2][v1] != -1) nattached++;
    if (nattached > 1) {
      if (nattached == 2) {
	if (complete[v0])
	  nattached = 1;
	else if (complete[v1])
	  nattached = 1;
	else if (complete[v2])
	  nattached = 1;
      }
      if (nattached > 1)
	if (!orientable()) {
	  remove_face();
	  return;
	}
    }
  }

  /* pruning after adding face */
  /* only upon completing a vertex */

  if (complete[v0]) 
    if (!min_lex()) {
      remove_face();
      return;
    }
  
  /* check for contractible edge */

  if (only_irreducible && nv > 4) {
    if (complete[v0])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v0][v] != -1)
	  if (contractible(v0,v)) {
	    remove_face();
	    return;
	  }
    if (complete[v1])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v1][v] != -1 &&
	    v != v0)
	  if (contractible(v1,v)) {
	    remove_face();
	    return;
	  }
    if (complete[v2])
      for (v=0; v<nv; v++)
	if (complete[v] &&
	    edge[v2][v] != -1 &&
	    v != v0 && v != v1)
	  if (contractible(v2,v)) {
	    remove_face();
	    return;
	  }
  }
  
  /* check surface is done */
  
  done = TRUE;
  for (v=v0; done && v<nv; v++)
    done = (complete[v]);

  if (done) {
    if (nf < minnf) {
      remove_face();
      return;
    }

    g = 2-nv+ne-nf;
    if (g < ming || g > maxg) {
      remove_face();
      return;
    }

    /* if orientability == 1 this has already been checked above */

    if (orientability == 1) 
      orient = TRUE;
    else {
      orient = orientable();
      if (orientability == 0 && orient) {
	remove_face();
	return;
      }
    }
    
    /* when checking for bipartite dual check now that any uncolored faces 
       can be properly colored. 
       this can happen if the dual was not connected at some point. */

    if (dual_bipartite) {
      uncolored = nf;
      for (iface=0; iface<nf; iface++) {
	v0 = list[iface][0];
	v1 = list[iface][1];
	v2 = list[iface][2];
	if (edge[v0][v1] == v2)
	  lcolor = color[v0][v1];
	else
	  lcolor = color[v1][v0];
	if (lcolor != -999)
	  uncolored--;
	fcolor[iface] = lcolor;
      }
      
      for (iface=0; iface<nf && uncolored > 0; iface++)
	if (fcolor[iface] == -999) {
	  v0 = list[iface][0];
	  v1 = list[iface][1];
	  v2 = list[iface][2];
	  
	  /* lcolor is the color of the neighbors of iface */
	  /* ocolor is the color of one (current) neighbor of iface */
	  
	  if (edge[v0][v1] == v2) 
	    ocolor = color[v1][v0];
	  else
	    ocolor = color[v0][v1];
	  lcolor = ocolor;
	  
	  if (edge[v1][v2] == v0) 
	    ocolor = color[v2][v1];
	  else
	    ocolor = color[v1][v2];
	  if (ocolor == -lcolor) {
	    remove_face();
	    return;
	  }
	  if (ocolor != -999)
	    lcolor = ocolor;
	  
	  if (edge[v2][v0] == v1) 
	    ocolor = color[v0][v2];
	  else
	    ocolor = color[v2][v0];
	  if (ocolor == -lcolor) {
	    remove_face();
	    return;
	  }
	  if (ocolor != -999)
	    lcolor = ocolor;
	  
	  if (lcolor != -999) {
	    fcolor[iface] = -lcolor;
	    uncolored--;
	  }
	}
    }
    
    if (nv >= minnv && nv <= maxnv) {
      found_one(g,orient);
    }
    
    /*
      dump_it();
    */
    
    remove_face();
    return;
  }

  /* do it again */

  next_face();

  remove_face();
}

void next_face() 
{
  /* a face has just been added.  find what the next face might be. */

  int v0,v1,v2;

  if (nf == splitlevel) {
#ifdef SPLITTEST
    ADDBIG(splitcases,1);
    return;
#endif
    if (splitcount-- != 0) return;
    splitcount = mod - 1;
  }

  v0 = list[nf-1][0];
  v1 = list[nf-1][1];
  v2 = list[nf-1][2]+1;

  /* check if v0 is complete */

  if (complete[v0]) {
    while (complete[v0])
      v0++;
    v1 = v0+1;
    v2 = v1+1;
  }

  /* while edge (v0,v1) is not on boundry */

  while ((edge[v0][v1] == -1 && edge[v1][v0] == -1) ||
	 (edge[v0][v1] != -1 && edge[v1][v0] != -1)) {
    if (edge[v0][v1] == -1 && 
	edge[v1][v0] == -1 &&
	!complete[v1])
      
      /* first use of the edge (v0,v1) */

      for (v2=v1+1; v2<MIN(nv+1,maxnv-1); v2++)
	if ((edge[v0][v2] == -1 || edge[v2][v0] == -1) &&
	    (edge[v1][v2] == -1 || edge[v2][v1] == -1) &&
	    (v2 == nv || !complete[v2]))
	  do_face(v0,v1,v2);
    v1++;
    v2 = v1+1;
  }

  /* edge (v0,v1) is on boundry.  must be used */
   
  for (; v2<nv+1; v2++)
    if ((edge[v0][v2] == -1 || edge[v2][v0] == -1) &&
	(edge[v1][v2] == -1 || edge[v2][v1] == -1) &&
	(v2 == nv || !complete[v2]))
      do_face(v0,v1,v2);
  
}
 
static void
initialize_splitting(int minlevel, int hint, int maxlevel)
  
/* Set splitlevel and splitcount.  */
{
  splitlevel = hint;
  if (splitlevel > maxlevel) splitlevel = maxlevel;
  
  if (splitlevel < minlevel && splitlevel > 0)
    {
      if (minlevel <= maxlevel) splitlevel = minlevel;
      else                      splitlevel = 0;
    }
  if (mod == 1) splitlevel = 0;
  splitcount = res;
}

static int 
getargvalue(char *arg)
  
/* Find integer value of arguement. 
   arg is a pointer to a command-line argument.
   The value of the arguement is the function return value.
   An absent value is equivalent to 0.
*/
{
  int j,ans,neg,tmp;
  
  ans = 0;
  if (arg[0] == '-') {
    neg = -1;
    j = 1;
  }
  else {
    neg = +1;
    j = 0;
  }
  
  for (; arg[j] >= '0' && arg[j] <= '9'; ++j) {
    tmp = ans;
    ans = ans * 10 + (arg[j] - '0');
    if (tmp != ans/10) {
      fprintf(stderr,"value too large.\n");
      return -999999;
    }
  }

  if (arg[j] != '\0')
    return -999999;
  else
    return neg*ans;
}

void summarize() 
{
  /* summarize results */

  int v,g;

  if (verbose) {
    PRINTBIG(stderr,ngenerated_total);
    fprintf(stderr," generated, total\n");
    for (v=3; v<=maxnv; v++)
      for (g=0; g<MAXG; g++)
	if (!ISZEROBIG(ngenerated[v][g][1])) {
	  PRINTBIG(stderr,ngenerated[v][g][1]);
	  fprintf(stderr,
		  " generated, %d vertices, Euler genus %d, orientable\n",
		  v,g);
	}
    for (v=3; v<=maxnv; v++)
      for (g=0; g<MAXG; g++)
	if (!ISZEROBIG(ngenerated[v][g][0])) {
	  PRINTBIG(stderr,ngenerated[v][g][0]);
	  fprintf(stderr,
		  " generated, %d vertices, Euler genus %d, nonorientable\n",
		  v,g);
	}
  }
  
}

int
main(argc,argv)
int argc;
char *argv[];

{
  int iarg;
  int badargs;
  char *arg;
  int hint;
  int degree0;
  int lg,lnv,lmindegree;
  int res_digits;
  char res_tmp[15];

#if CPUTIME
  struct tms timestruct0,timestruct1;
  
  times(&timestruct0);
#endif
 
  only_irreducible = FALSE;
  even_degree = FALSE;
  dual_bipartite = FALSE;
  vertex_transitive = FALSE;
  orientability = -1;
  equivelar = -1;
  mindegree = -1;
  maxdegree = MAXN-1;
  minnv = 4;
  maxnv = MAXN;
  minnf = 4;
  maxnf = MAXF;
  ming = 0;
  maxg = MAXG;
  pq_switch = FALSE;
  res = 0;
  mod = 1;
  badargs = FALSE;

  for (iarg = 0; iarg < argc ; ++iarg) fprintf(stderr,"%s ",argv[iarg]);
  fprintf(stderr,"\n");

  if (argc < 2) {
    fprintf(stderr,"%s\n",HELPTEXT);
    exit(1);
  }

  iarg = 1;
  while (iarg < argc) {
    arg = argv[iarg];
    if (strcmp(arg,"-v") == 0) {
      verbose = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-h") == 0) {
      fprintf(stderr,"%s\n",HELPTEXT);
      exit(1);
    }
    else if (strcmp(arg,"-r") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"res value required after -r switch\n");
	badargs = TRUE;
      }
      else {
	res = getargvalue(argv[iarg]);
	iarg++;
	if (res < 0) {
	  fprintf(stderr,"res (%d) must be nonnegative.\n",res);
	  badargs = TRUE;
	  res = 0;
	}
      }
    }
    else if (strcmp(arg,"-m") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"mod value required after -m switch\n");
	badargs = TRUE;
      }
      else {
	mod = getargvalue(argv[iarg]);
	iarg++;
	if (mod < 1) {
	  fprintf(stderr,"mod (%d) must be posotive.\n",mod);
	  badargs = TRUE;
	  mod = 1;
	}
      }
    }
    else if (strcmp(arg,"-o") == 0) {
      if (orientability == 0 || orientability == 1) {
	fprintf(stderr,"o switch is redefining orientability\n");
	badargs = TRUE;
      }
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -o switch\n");
	badargs = TRUE;
      }
      else {
	orientability = getargvalue(argv[iarg]);
	iarg++;
	if (orientability != 0 && orientability != 1) {
	  fprintf(stderr,"orientability (%d) must be 0 or 1.\n",
		  orientability);
	  badargs = TRUE;
	  orientability = -1;
	}
      }
    }
    else if (strcmp(arg,"-e") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -e switch\n");
	badargs = TRUE;
      }
      else {
	equivelar = getargvalue(argv[iarg]);
	iarg++;
	if (equivelar < 3) {
	  fprintf(stderr,"equivelar degree must be at least 3\n");
	  badargs = TRUE;
	  equivelar = 3;
	}
      }
    }
    else if (strcmp(arg,"-d") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -d switch\n");
	badargs = TRUE;
      }
      else {
	mindegree = getargvalue(argv[iarg]);
	iarg++;
	if (mindegree < 3) {
	  fprintf(stderr,"minimum degree must be at least 3\n");
	  badargs = TRUE;
	  mindegree = 3;
	}
      }
    }
    else if (strcmp(arg,"-D") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -D switch\n");
	badargs = TRUE;
      }
      else {
	maxdegree = getargvalue(argv[iarg]);
	iarg++;
	if (maxdegree > MAXN-1) {
	  fprintf(stderr,"maximum degree must be at most %d\n",MAXN-1);
	  badargs = TRUE;
	  mindegree = MAXN-1;
	}
      }
    }
    else if (strcmp(arg,"-g") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -g switch\n");
	badargs = TRUE;
      }
      else {
	ming = getargvalue(argv[iarg]);
	iarg++;
      }
    }
    else if (strcmp(arg,"-G") == 0) {
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"value required after -G switch\n");
	badargs = TRUE;
      }
      else {
	maxg = getargvalue(argv[iarg]);
	iarg++;
      }
    }
    else if (strcmp(arg,"-i") == 0) {
      only_irreducible = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-E") == 0) {
      even_degree = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-B") == 0) {
      dual_bipartite = TRUE;
      even_degree = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-T") == 0) {
      vertex_switch = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-F") == 0) {
      flag_switch = TRUE;
      iarg++;
    }
    else if (strcmp(arg,"-pq") == 0) {
      pq_switch = TRUE;
      iarg++;
      if (iarg >= argc) {
	fprintf(stderr,"2 values required after -pq switch\n");
	badargs = TRUE;
      }
      else {
	pq_degrees[0] = getargvalue(argv[iarg]);
	if (pq_degrees[0] < 3) {
	  fprintf(stderr,"value (%d) with -pq switch too small\n",
		  pq_degrees[0]);
	  badargs = TRUE;
	}
	iarg++;
	if (iarg >= argc) {
	  fprintf(stderr,"2 values required after -pq switch\n");
	  badargs = TRUE;
	}
	else {
	  pq_degrees[1] = getargvalue(argv[iarg]);
	  if (pq_degrees[1] < 3) {
	    fprintf(stderr,"value (%d) with -pq switch too small\n",
		    pq_degrees[1]);
	    badargs = TRUE;
	  }
	  pq_degrees[1] = 2*pq_degrees[1];
	  iarg++;
	}
      }
      mindegree = MIN(pq_degrees[0],pq_degrees[1]);
      maxdegree = MAX(pq_degrees[0],pq_degrees[1]);
      pq_total = pq_degrees[0]+2*pq_degrees[1];
    }
    else if (iarg < argc-1) {
      iarg++;
      fprintf(stderr,"unknown switch %s\n",arg);
      badargs = TRUE;
    }
    else {
      maxnv = getargvalue(arg);
      iarg++;
      if (maxnv == -999999) {
	fprintf(stderr,"unknown switch %s\n",arg);
	badargs = TRUE;
      }
    }
  }

  if (equivelar < 3 && !pq_switch)
    minnv = maxnv;

  if (maxnv > MAXN) {
    fprintf(stderr,"maxnv (%d) too large.  increase MAXN.\n",maxnv);
    badargs = TRUE;
  }

  if (minnv < 4) {
    fprintf(stderr,"minnv (%d) too small.\n",minnv);
    badargs = TRUE;
  }

  if (minnv > maxnv) {
    fprintf(stderr,"minnv (%d) > maxnv (%d).\n",minnv,maxnv);
    badargs = TRUE;
  }

  if (maxg > MAXG) {
    fprintf(stderr,"maxg (%d) too large.  increase MAXG.\n",maxg);
    badargs = TRUE;
  }

  if (ming < 0) {
    fprintf(stderr,"ming (%d) too small.\n",ming);
    badargs = TRUE;
  }

  if (ming > maxg) {
    fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
    badargs = TRUE;
  }

  if (maxg > (maxnv-4)*(maxnv-3)/6) {
    fprintf(stderr,"maxg (%d) is not obtainable for specified maxnv (%d).\n",
	    maxg,maxnv);
    maxg = (maxnv-4)*(maxnv-3)/6;
    fprintf(stderr,"reducing maxg to %d.\n",maxg);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (orientability == 1 && maxg%2 != 0) {
    fprintf(stderr,"Only orientable triangulations and maxg (%d) is odd.\n",
	    maxg);
    maxg--;
    fprintf(stderr,"reducing maxg to %d.\n",maxg);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (orientability == 1 && ming%2 != 0) {
    fprintf(stderr,"Only orientable triangulations and ming (%d) is odd.\n",
	    ming);
    ming++;
    fprintf(stderr,"increasing ming to %d.\n",ming);
    if (ming > maxg) {
      fprintf(stderr,"ming (%d) > maxg (%d).\n",ming,maxg);
      badargs = TRUE;
    }
  }

  if (ming > (minnv-4)*(minnv-3)/6) {
    fprintf(stderr,"ming (%d) is not obtainable for specified minnv (%d).\n",
	    ming,minnv);
    while (ming > (minnv-4)*(minnv-3)/6)
      minnv++;
    fprintf(stderr,"increasing minnv to %d.\n",minnv);
    if (minnv > maxnv) {
      fprintf(stderr,"minnv (%d) > maxnv (%d).\n",minnv,maxnv);
      badargs = TRUE;
    }
  }
  
  minnf = 2*(ming-2+minnv);
  maxnf = 2*(maxg-2+maxnv);

  if (maxnf >= MAXF) {
    fprintf(stderr,"number of faces (%d) too large.  increase MAXF.\n",
	    maxnf);
    badargs = TRUE;
  }

  if (mindegree == -1) {
    mindegree = maxnv;
    
    for (lg=ming; lg<=maxg; lg++)
      for (lnv=minnv; lnv<=maxnv; lnv++) {
	lmindegree = MAX(3,6*(lg-2+lnv)-(lnv-1)*(lnv-1));
	if (lmindegree < mindegree)
	  mindegree = lmindegree;
      }
  }

  if (maxdegree >= maxnv) {
    if (maxdegree != MAXN-1) {
      fprintf(stderr,
	      "maxdegree (%d) must be less than number of vertices (%d).\n",
	      maxdegree,maxnv);
      badargs = TRUE;
    }
    maxdegree = maxnv-1;
  }

  if (even_degree && !equivelar) {
    if (mindegree%2 != 0) {
      fprintf(stderr,
	      "since all vertices have even degree mindegree (%d) must be even (%d) also.\n",
	      mindegree,mindegree+1);
      mindegree++;
    }
    if (maxdegree%2 != 0) {
      fprintf(stderr,
	      "since all vertices have even degree maxdegree (%d) must be even (%d) also.\n",
	      maxdegree,maxdegree-1);
      maxdegree--;
    }
  }

  if (mindegree > maxdegree) {
      fprintf(stderr,
	      "mindegree (%d) must be at most maxdegree (%d).\n",
	      mindegree,maxdegree);
      badargs = TRUE;
    }

  if (vertex_switch && flag_switch) {
    fprintf(stderr,
	      "-T and -F switches are incompatible.\n");
    badargs = TRUE;
  }

  if (equivelar > 0) {
    if (mindegree > equivelar || maxdegree < equivelar) {
      fprintf(stderr,
	      "equivelar (%d) not between mindegree (%d) and maxdegree (%d).\n",
	      equivelar,mindegree,maxdegree);
      badargs = TRUE;
    }
    else {
      mindegree = equivelar;
      maxdegree = equivelar;
    }
  }
  else {
    if (vertex_switch && !pq_switch) {
      fprintf(stderr,
	      "-e or -pq must be specified with -T.\n");
      badargs = TRUE;
    }
    if (vertex_switch && pq_switch)
      if (pq_degrees[1] > pq_degrees[0]) {
	fprintf(stderr,
		"with -T and -pq switches 2q > p is not implemented.\n");
	badargs = TRUE;
      }
    if (flag_switch && !pq_switch) {
      fprintf(stderr,
	      "-e or -pq must be specified with -F.\n");
      badargs = TRUE;
    }
  }

  if (res >= mod) {
    fprintf(stderr,"res (%d) >= mod (%d).\n",res,mod);
    badargs = TRUE;
    res = 0;
    mod = 1;
  }

  if (badargs) {
    fprintf(stderr,"Usage: %s\n",USAGE);
    exit(1);
  }
  
#ifdef SPLITTEST
  if (mod == 1) mod = 2;
#endif

  if (verbose) {
    fprintf(stderr,"Minimum number of vertices: %d\n", minnv);
    fprintf(stderr,"Maximum number of vertices: %d\n", maxnv);
    fprintf(stderr,"Minimum number of faces: %d\n", minnf);
    fprintf(stderr,"Maximum number of faces: %d\n", maxnf);
    fprintf(stderr,"Minimum Euler genus: %d\n", ming);
    fprintf(stderr,"Maximum Euler genus: %d\n", maxg);
    fprintf(stderr,"Minimum degree of vertices: %d\n", mindegree);
    fprintf(stderr,"Maximum degree of vertices: %d\n", maxdegree);
    if (orientability == 0)
      fprintf(stderr,"Only nonorientable are generated\n");
    else if (orientability == 1)
      fprintf(stderr,"Only orientable are generated\n");
    else
      fprintf(stderr,"Either nonorientable or orientable are generated\n");
    if (only_irreducible)
      fprintf(stderr,"Only irreducible triangulations are generated.\n");
    if (even_degree)
      fprintf(stderr,"Every vertex has even degree.\n");
    if (dual_bipartite)
      fprintf(stderr,"Dual of triangulation is bipartite.\n");
    if (flag_switch)
      fprintf(stderr,"Only flag transitive triangulations are generated.\n");
    else if (vertex_switch)
      fprintf(stderr,"Only vertex transitive triangulations are generated.\n");
    if (pq_switch)
      fprintf(stderr,"Triangulations of equivelar polyhedral maps of type {%d,%d}.\n",
	      pq_degrees[0],pq_degrees[1]/2);
    fprintf(stderr,"res/mod = %d/%d\n",res,mod);
  }

  hint = MIN(maxnf/2, (3*maxnv)/2);
  if (maxnv <= 16)
    hint = MIN(hint, maxnv+7);
  else if (maxnv <= 19)
    hint = MIN(hint, 24);
  else if (maxnv <= 50)
    hint = MIN(hint, maxnv+4);
  else
    hint = MIN(hint, 54);

  initialize_splitting(4,hint,maxnf-1);

  if (verbose && mod != 1)
    fprintf(stderr,"splitlevel is at %d faces.\n",splitlevel);

  sprintf(res_tmp,"%d",mod-1);
  res_digits = strlen(res_tmp);
  sprintf(res_tmp,"%c%d.%dd",'%',res_digits,res_digits);
  sprintf(res_text,res_tmp,res);

  initialize();

  for (degree0=mindegree; degree0 <= maxdegree; degree0++) {
    if (!pq_switch) {
      if (!even_degree || degree0%2 == 0) {
	initialize_first_vertex(degree0,-1,-1);
	next_face();
      }
    }
    else {
      if (degree0 == pq_degrees[0] && degree0 == pq_degrees[1]) {
	initialize_first_vertex(degree0,1,1);
	next_face();
      }
      else if (degree0 == pq_degrees[0] && degree0 < pq_degrees[1]) {
	initialize_first_vertex(degree0,0,1);
	next_face();
      }
      else if (degree0 == pq_degrees[1] && degree0 < pq_degrees[0]) {
	initialize_first_vertex(degree0,1,0);
	next_face();
	initialize_first_vertex(degree0,1,1);
	next_face();
      }
    }
  }

#if CPUTIME
  times(&timestruct1);
#endif
  
#ifdef SPLITTEST
  PRINTBIG(stderr,splitcases);
  fprintf(stderr," splitting cases at level = %d",splitlevel);
#if CPUTIME
  fprintf(stderr,"; cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime-timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif
  return 0;
#endif
  
  summarize();
  
#if CPUTIME
  fprintf(stderr,"cpu=%.2f sec\n",
	  (double)(timestruct1.tms_utime+timestruct1.tms_stime
		   -timestruct0.tms_utime+timestruct0.tms_stime) / 
	  (double)CLK_TCK);
#else
  fprintf(stderr,"\n");
#endif

  exit(0);
}
